from typing import Any

from pydantic_core import CoreSchema
from pydantic_core import core_schema
from pydantic.json_schema import JsonSchemaValue
from pydantic import GetJsonSchemaHandler
from libcanonical.utils.encoding import b64encode_int
from libcanonical.utils.encoding import b64decode_int


__all__: list[str] = [
    'Base64Int'
]


class Base64Int(int):
    __module__: str = 'libcanonical.types'

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=core_schema.str_schema(),
            python_schema=core_schema.chain_schema([
                core_schema.no_info_plain_validator_function(cls.b64decode)
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(
                cls.b64encode
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.str_schema())

    @classmethod
    def b64decode(cls, value: bytes | str | int):
        if not isinstance(value, int):
            value = b64decode_int(value)
            raise Exception
        return value

    @classmethod
    def b64encode(cls, value: int) -> str:
        return  bytes.decode(b64encode_int(value), 'ascii')


from typing import Any
from typing import Optional
from typing import TypedDict


LoggerConfigDict = TypedDict('LoggerConfigDict', {
    'handlers': list[str],
    'level': str,
    'propagate': bool
})


class LoggingConfigDict(TypedDict):
    version: int
    disable_existing_loggers: bool
    filters: Optional[dict[str, Any]]
    loggers: dict[str, LoggerConfigDict]


import inspect
import operator
from collections import abc
from collections import OrderedDict
from functools import reduce
from itertools import chain
from typing import Any
from typing import Callable

from ._chunks import chunks
from ._delay import delay
from ._deephash import deephash
from ._logging import logger
from ._retry import retry
from .deferred import Deferred
from .loader import import_symbol


__all__: list[str] = [
    'chunks',
    'deephash',
    'delay',
    'import_symbol',
    'logger',
    'retry',
    'Deferred',
]


class class_property:
    __module__: str = 'oauthx.utils'

    def __init__(self, func: Callable[..., Any]):
        self.func = func

    def __get__(self, instance: Any, cls: Any) -> Any:
        return self.func(cls)


def merge_signatures(
    signatures: list[inspect.Signature],
    extra: list[inspect.Parameter] | None = None
) -> inspect.Signature:
    """Merge signatures to that FastAPI can inject the dependencies."""
    extra = extra or []
    params: dict[str, inspect.Parameter] = OrderedDict()
    for param in chain(chain(*[x.parameters.values() for x in signatures]), extra):
        if param.name in {'self', 'cls'}:
            continue
        if param.name.startswith('_'):
            continue
        if param.kind == inspect.Parameter.VAR_POSITIONAL:
            continue
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            continue
        params[param.name] = param

    return signatures[0].replace(
        parameters=list(sorted(params.values(), key=lambda p: (p.kind, p.default != inspect._empty))) # type: ignore
    )


def merged_call(func: Callable[..., Any], kwargs: Any) -> Any:
    sig = inspect.signature(func)
    return func(**{k: v for k, v in kwargs.items() if k in sig.parameters})


def throw(cls: type[Exception], *args: Any, **kwargs: Any):
    raise cls(*args, **kwargs)


def jsonpath(mapping: abc.Mapping[str, Any], path: str) -> Any:
    if not str.startswith(path, '/'):
        raise ValueError("Path must start with a slash.")
    parts: list[str] = []
    for p in str.split(str.lstrip(path, '/'), '/'):
        p = p.replace('~0', '~')
        p = p.replace('~1', '/')
        parts.append(p)
    return reduce(operator.getitem, parts, mapping)
    


import pydantic


class Request(pydantic.BaseModel):
    host: str = pydantic.Field(
        default=...,
        description=(
            "The IP address of the client, as seen by the "
            "server."
        )
    )

    url: str = pydantic.Field(
        default=...,
        description="The full URL that was requested."
    )


import hashlib

from .base64 import Base64

__all__: list[str] = [
    'DigestSHA256'
]


class DigestSHA256(Base64):
    __module__: str = 'libcanonical.types'
    dig = 'sha256'
    max_length = 32
    min_length = 32

    @classmethod
    def hasher(cls):
        return hashlib.new(cls.dig)


from .applicationruntimestate import ApplicationRuntimeState
from .awaitablebool import AwaitableBool
from .awaitablebytes import AwaitableBytes
from .base64 import Base64
from .base64int import Base64Int
from .base64json import Base64JSON
from .base64urlencoded import Base64URLEncoded
from .bcp47 import BCP47
from .bytestype import BytesType
from .colonseparatedlist import ColonSeparatedList
from .colonseparatedset import ColonSeparatedSet
from .crypto import EncryptionResult
from .digestsha256 import DigestSHA256
from .domainname import DomainName
from .emailaddress import EmailAddress
from .exceptions import *
from .hexencoded import HexEncoded
from .httpretryafter import HTTPRetryAfter
from .httprequestref import HTTPRequestRef
from .httpresourcelocator import HTTPResourceLocator
from .jsonpath import JSONPath
from .phonenumber import Phonenumber
from .pythonsymbol import PythonSymbol
from .resourcename import ResourceName
from .resourcename import TypedResourceName
from .serializableset import SerializableSet
from .stringorset import StringOrSet
from .stringtype import StringType
from .unixtimestamp import UnixTimestamp
from .websocketresourcelocator import WebSocketResourceLocator


__all__: list[str] = [
    'ApplicationRuntimeState',
    'AwaitableBool',
    'AwaitableBytes',
    'Base64',
    'Base64Int',
    'Base64JSON',
    'Base64URLEncoded',
    'BytesType',
    'BCP47',
    'ColonSeparatedList',
    'ColonSeparatedSet',
    'Conflict',
    'DigestSHA256',
    'DomainName',
    'EmailAddress',
    'EncryptionResult',
    'ExceptionRaiser',
    'HexEncoded',
    'HTTPRetryAfter',
    'HTTPRequestRef',
    'HTTPResourceLocator',
    'JSONPath',
    'Phonenumber',
    'PythonSymbol',
    'ResourceName',
    'SerializableSet',
    'StringOrSet',
    'StringType',
    'TypedResourceName',
    'Undecryptable',
    'UnixTimestamp',
    'WebSocketResourceLocator',
]


from ._externalstate import ExternalState


class PollingExternalState(ExternalState):
    """An :class:`~libcanonical.bases.ExternalState` implementation that
    uses polling as it's transport mechanism.
    """
    __module__: str = 'libcanonical.bases'

    #: The interval between queries at the remote source for the
    #: current state.
    interval: float = 10.0

    #

    async def poll(self) -> None:
        """Polls the remote source for data."""
        raise NotImplementedError


import sys


class FatalException(SystemExit):
    """Represent an exception condition which prevents an process from
    continuing to run.
    """
    exitcode: int = 1

    def __init__(self, reason: str):
        sys.stderr.write(reason + '\n')
        sys.stderr.flush()
        super().__init__(self.exitcode)


import pydantic
import pytest

from libcanonical.types import HTTPRetryAfter


@pytest.mark.parametrize("value", [
    "2",
    "Wed, 23 Sep 2009 22:15:29 GMT"
])
def test_valid_values(value: str):
    HTTPRetryAfter.parse(value)


@pytest.mark.parametrize("value", [
    "a",
])
def test_invalid_values(value: str):
    header = HTTPRetryAfter.parse(value)
    assert header.seconds == HTTPRetryAfter.default_delay


@pytest.mark.parametrize("value", [
    None,
])
def test_none(value: str):
    header = HTTPRetryAfter.parse(value)
    assert header.seconds == 0.0




class AwaitableBool(int):
    """A bool that can be awaited. For sync/async cross-compatibility."""

    async def _async(self):
        return bool(self)

    def __await__(self):
        return self._async().__await__()


import pydantic

from libcanonical.types.base64 import Base64


class BaseEncryptionResult(pydantic.BaseModel):
    alg: str = pydantic.Field(
        default=...,
        title="Algorithm",
        description="The algorithm used to perform the encryption function."
    )

    kid: str | None = pydantic.Field(
        default=None,
        title="Key Identifier",
        description=(
            "Identifies the key used to perform the encryption function."
        )
    )

    ct: Base64 = pydantic.Field(
        default=...,
        title="Ciphertext",
        description=(
            "The Base64-encoded encryption result."
        )
    )

    def __bytes__(self) -> bytes:
        return self.ct


from typing import Any
from typing import ClassVar
from typing import Self
from typing import TypeVar

import pydantic
from pydantic_core import CoreSchema
from pydantic_core import core_schema
from pydantic.json_schema import JsonSchemaValue
from pydantic import GetJsonSchemaHandler


T = TypeVar('T')


class HTTPHeaderType:
    __module__: str = 'libcanonical.types'
    openapi_title: ClassVar[str | None] = None
    openapi_description: ClassVar[str | None] = None
    max_length: int | None = None
    min_length: int | None = None

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=cls.__default_schema__(),
            python_schema=core_schema.union_schema([
                core_schema.is_instance_schema(cls),
                core_schema.chain_schema([
                    cls.__default_schema__(),
                    core_schema.no_info_plain_validator_function(cls.transform),
                    core_schema.no_info_plain_validator_function(cls.validate),
                    core_schema.no_info_plain_validator_function(cls)
                ])
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(str)
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        schema = handler(cls.__default_schema__())
        schema['title'] = cls.openapi_title or cls.__name__
        if cls.openapi_title is not None:
            schema['description'] = cls.openapi_title
        return schema

    @classmethod
    def __default_schema__(cls):
        return core_schema.str_schema(
            max_length=cls.max_length,
            min_length=cls.min_length,
        )

    @classmethod
    def transform(cls, v: str) -> str:
        return v

    @classmethod
    def parse(cls, value: str) -> Self:
        adapter: pydantic.TypeAdapter[Any] = pydantic.TypeAdapter(cls)
        return adapter.validate_python(value)

    @classmethod
    def validate(cls, v: str) -> str:
        return v

    def __str__(self) -> str:
        raise NotImplementedError

    def __repr__(self) -> str: # pragma: no cover
        return f'<{type(self).__name__}: {str(self)}>'


from ._base import BaseCanonicalException


class Undecryptable(BaseCanonicalException):
    """This exception may be raised when the decryption of a ciphertext
    is attempted but the operation failed.
    """
    http_status_code: int = 503


import asyncio
import datetime
import logging
import email.utils
from typing import TypeVar

from pydantic_core import core_schema

from .httpheadertype import HTTPHeaderType


T = TypeVar('T')

INVALID = object()


class HTTPRetryAfter(HTTPHeaderType):
    __module__: str = 'libcanonical.types'
    default_delay: float = 5.0
    nbf: datetime.datetime
    logger: logging.Logger = logging.getLogger(__name__)

    @property
    def seconds(self):
        return int((self.nbf - self.now).total_seconds())

    @classmethod
    def __default_schema__(cls):
        return core_schema.union_schema([
            core_schema.chain_schema([
                core_schema.no_info_plain_validator_function(email.utils.parsedate_to_datetime),
                core_schema.no_info_plain_validator_function(
                    lambda x: (
                        datetime.datetime.astimezone(x, datetime.timezone.utc)
                    )
                )
            ]),
            core_schema.chain_schema([
                core_schema.int_schema(ge=0),
                core_schema.no_info_plain_validator_function(lambda x: (
                    datetime.datetime.now(datetime.timezone.utc)
                    + datetime.timedelta(seconds=x)
                ))
            ]),
            core_schema.chain_schema([
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(
                    lambda x: cls.logger.warning('Invalid Retry-After value: %s', str(x)[:32])
                ),
                core_schema.no_info_plain_validator_function(lambda x: INVALID)
            ]),
            core_schema.none_schema(),
        ])

    def __init__(self, nbf: datetime.datetime | None):
        self.now = datetime.datetime.now(datetime.timezone.utc)
        if nbf == INVALID:
            nbf = self.now + datetime.timedelta(seconds=5)
        self.nbf = nbf or self.now
        assert isinstance(self.nbf, datetime.datetime)

    def delay(
        self,
        default: float | None = None,
        max_delay: float | None = None
    ):
        num_seconds = max(default or self.seconds, self.seconds)
        return asyncio.sleep(min(num_seconds, max_delay or num_seconds))

    def __str__(self): # pragma: no cover
        return str(self.nbf)


from typing import overload
from typing import Awaitable
from typing import Literal
from typing import Protocol


class ISigner(Protocol):

    @overload
    def sign(
        self,
        message: bytes | str
    ) -> Awaitable[bytes]:
        ...

    @overload
    def sign(
        self,
        message: bytes | str,
        *,
        blocking: Literal[False]
    ) -> Awaitable[bytes]:
        ...

    @overload
    def sign(
        self,
        message: bytes | str,
        *,
        blocking: Literal[True]
    ) -> bytes:
        ...

    def sign(
        self,
        message: bytes | str,
        *,
        blocking: bool = False
    ) -> bytes | Awaitable[bytes]:
        ...


from typing import Union

import pydantic

from ._aesencryptionresult import AESEncryptionResult
from ._bytesencryptionresult import BytesEncryptionResult
from ._dhencryptionresult import DHEncryptionResult


EncryptionResultType = Union[
    AESEncryptionResult,
    BytesEncryptionResult,
    DHEncryptionResult
]
    

class EncryptionResult(pydantic.RootModel[EncryptionResultType]):

    @property
    def aad(self) -> bytes:
        return self.root.aad

    @property
    def iv(self) -> bytes | None:
        return self.root.iv

    @property
    def tag(self) -> bytes | None:
        return self.root.tag

    def __bytes__(self) -> bytes:
        return bytes(self.root)


from typing import Any
from typing import Callable
from typing import TypeVar

import phonenumbers
from pydantic import GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import CoreSchema
from pydantic_core import core_schema


T = TypeVar('T')


class Phonenumber(str):

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=core_schema.str_schema(max_length=128),
            python_schema=core_schema.union_schema([
                core_schema.is_instance_schema(cls),
                core_schema.chain_schema([
                    core_schema.str_schema(max_length=128),
                    core_schema.no_info_plain_validator_function(cls.fromstring)
                ])
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(str)
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        schema = handler(core_schema.str_schema(max_length=18))
        schema['title'] = 'Phonenumber'
        schema['description'] = (
            "The ITU-T phone number format, defined by E.164, is an international "
            "numbering standard that ensures a globally unique structure for "
            "telephone numbers, consisting of a country code (1–3 digits), a "
            "national destination code (NDC), and a subscriber number, with a "
            "maximum length of 15 digits."
        )
        return schema

    @property
    def maskable(self) -> bytes:
        return str.encode(f'phonenumber:{self.lower()}', 'ascii')

    @classmethod
    def fromstring(cls, v: Any, _: Any = None) -> str:
        if not isinstance(v, str):
            raise TypeError("string required")
        try:
            p = phonenumbers.parse(v)
            if not phonenumbers.is_valid_number(p):
                raise ValueError("Not a valid phonenumber.")
        except (phonenumbers.NumberParseException, TypeError):
            raise ValueError("Not a valid phonenumber.")
        return cls(
            phonenumbers.format_number(p, phonenumbers.PhoneNumberFormat.E164)
        )

    @classmethod
    def parse(
        cls,
        v: str,
        region: str | None = None
    ):
        try:
            p = phonenumbers.parse(v, region=region)
            if not phonenumbers.is_valid_number(p):
                raise ValueError("Not a valid phonenumber.")
        except (phonenumbers.NumberParseException, TypeError):
            raise ValueError("Not a valid phonenumber.")
        return cls(
            phonenumbers.format_number(p, phonenumbers.PhoneNumberFormat.E164)
        )

    def mask(self, masker: Callable[[bytes], T]) -> T:
        return masker(self.maskable)

    def __repr__(self) -> str: # pragma: no cover
        return f'<Phonenumber: {str(self)}>'


import inspect
from typing import Any
from typing import Coroutine
from types import TracebackType


class Lifecycled:
    __module__: str = 'libcanonical.bases'

    def setup(
        self,
        reloading: bool = False
    ) -> None | Coroutine[Any, Any, None]:
        raise NotImplementedError

    def teardown(
        self,
        exception: BaseException | None = None
    ) -> bool | Coroutine[Any, Any, bool]:
        raise NotImplementedError

    async def __aenter__(self):
        result = self.setup(reloading=False)
        if inspect.isawaitable(result):
            await result
        return self

    async def __aexit__(
        self,
        cls: type[BaseException] | None = None,
        exception: BaseException | None = None,
        traceback: TracebackType | None = None
    ) -> bool:
        result = self.teardown()
        if inspect.isawaitable(result):
            await result
        assert isinstance(result, bool)
        return result


from .formatters import *
from ._loggingconfigdict import LoggingConfigDict


__all__: list[str] = [
    'AccessFormatter',
    'ColourizedFormatter',
    'DefaultFormatter',
    'JSONFormatter',
    'LoggingConfigDict',
]


import pydantic
import pytest

from canonical import HTTPResourceLocator



class Model(pydantic.BaseModel):
    url: HTTPResourceLocator


@pytest.mark.parametrize('value,exc_class', [
    ('http://www.google.com', None),
    ('https://www.google.com', None),
    ('file://www.google.com', pydantic.ValidationError),
    ('+31456', pydantic.ValidationError),
    ('a', pydantic.ValidationError),
    (1, pydantic.ValidationError)
])
def test_validate(value: str, exc_class: type[BaseException] | None):
    adapter = pydantic.TypeAdapter(HTTPResourceLocator)
    if exc_class is None:
        adapter.validate_python(value)
    else:
        with pytest.raises(exc_class):
            adapter.validate_python(value)


@pytest.mark.parametrize('value', [
    'http://www.google.com',
    'https://www.google.com',
])
def test_validate_model(value: str):
    obj = Model.model_validate({'url': value})
    assert isinstance(obj.url, HTTPResourceLocator)